(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-Clang                                      *)
(*                                                                        *)
(*  Copyright (C) 2012-2017                                               *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file LICENSE).                      *)
(*                                                                        *)
(**************************************************************************)

open Intermediate_format
open Genlex

let dkey = Frama_Clang_option.register_category "parse_ident"

let builtin_types = [ "__builtin_va_list" ]

let is_builtin_type t = List.mem t builtin_types

let is_builtin_qual_type n = List.mem n.decl_name builtin_types

let add_qualifier qual typ =
  if List.mem qual typ.qualifier then typ else
    { typ with qualifier = qual :: typ.qualifier }

let lexer = Genlex.make_lexer
  [ "("; ")"; "::"; "<"; ">"; ","; "*"; "&"; "&&"; "~";
    "bool"; "char"; "const"; "float"; "double"; "int"; "long"; "short";
    "static"; "unsigned"; "void"; "volatile"; "wchar_t";
  ]

let identify_u_long b = parser
(*    [< 'Kwd "long"; 'Kwd "int" >] -> ST_UNSIGNED_LONG_LONG *)
  | [< 'Kwd "int" >] -> 
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int IULong
  | [< >] -> Intermediate_format.Int IULong

let identify_u_short b = parser
  | [< 'Kwd "int" >] -> 
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int IUShort
  | [< >] -> Intermediate_format.Int IUShort

let identify_s_long b = parser
(*    [< 'Kwd "long"; 'Kwd "int" >] -> ST_UNSIGNED_LONG_LONG *)
  | [< 'Kwd "int" >] -> 
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int ILong
  | [< >] -> Intermediate_format.Int ILong

let identify_s_short b = parser
  | [< 'Kwd "int" >] -> 
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int IShort
  | [< >] -> Intermediate_format.Int IShort

let rec identify_unsigned b = parser
    [< 'Kwd "long"; s >] -> identify_u_long b s
  | [< 'Kwd "short"; s>] -> identify_u_short b s
  | [< 'Kwd "int"; s >] ->
      if b then raise (Stream.Error "illegal type specification")
      else identify_unsigned true s
  | [< 'Kwd "char" >] ->
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int IUChar
  | [< >] -> Intermediate_format.Int IUInt

let rec identify_signed b = parser
    [< 'Kwd "long"; s >] -> identify_s_long b s
  | [< 'Kwd "short"; s>] -> identify_s_short b s
  | [< 'Kwd "int"; s >] ->
      if b then raise (Stream.Error "illegal type specification")
      else identify_signed true s
  | [< 'Kwd "char" >] ->
      if b then raise (Stream.Error "illegal type specification")
      else Intermediate_format.Int ISChar
  | [< >] -> Intermediate_format.Int IInt

let identify_long_double = parser
(*    [< 'IDENT "_Complex" >] -> ST_LONG_DOUBLE_COMPLEX
  | [< 'IDENT "_Imaginary">] -> ST_LONG_DOUBLE_IMAGINARY *)
  | [< >] -> Intermediate_format.Float FLongDouble

let identify_float = parser
(*     [< 'IDENT "_Complex" >] -> ST_FLOAT_COMPLEX
  | [< 'IDENT "_Imaginary">] -> ST_FLOAT_IMAGINARY *)
  | [< >] -> Intermediate_format.Float FFloat

let rec identify_long b = parser
  | [< 'Kwd "int"; s >] ->
      if b then raise (Stream.Error "illegal type specification")
      else identify_long true s
  | [< 'Kwd "signed" ; s>] -> identify_s_long b s
  | [< 'Kwd "unsigned" ; s>] -> identify_u_long b s
  (* | [< 'IDENT "long"; 'IDENT "int" >] -> ST_LONG_LONG *)
  | [< 'Kwd "double"; s >] ->
      if b then raise (Stream.Error "illegal type specification")
      else identify_long_double s
  | [< >] -> Intermediate_format.Int ILong

let rec identify_short b = parser
  | [< 'Kwd "int"; s >] ->
      if b then raise (Stream.Error "illegal type specification")
      else identify_short true s
  | [< 'Kwd "signed" ; s>] -> identify_s_short b s
  | [< 'Kwd "unsigned" ; s>] -> identify_u_short b s
  
let identify_double = parser
(*    [< 'IDENT "_Complex" >] -> ST_DOUBLE_COMPLEX
  | [< 'IDENT "_Imaginary">] -> ST_DOUBLE_IMAGINARY *)
  | [< 'Kwd "long"; s >] -> identify_long_double s
  | [< >] -> Intermediate_format.Float FDouble

let identify_char = parser
  | [< 'Kwd "signed" >] -> Intermediate_format.Int ISChar
  | [< 'Kwd "unsigned" >] -> Intermediate_format.Int IUChar
  | [< >] -> Intermediate_format.Int IChar

let identify_int = parser
  | [< 'Kwd "unsigned"; s >] -> identify_unsigned true s
  | [< 'Kwd "long"; s >] -> identify_long true s
  | [< 'Kwd "short"; s>] -> identify_short true s
  | [< 'Kwd "signed"; s >] -> identify_signed true s
  | [< >] -> Intermediate_format.Int IInt

let decl_flag = parser
  | [< 'Kwd "const" >] -> Const
  | [< 'Kwd "volatile" >] -> Volatile
  | [< 'Kwd "restrict" >] -> Restrict
  | [< 'Kwd "static" >] -> Static

let opt_dimension = parser
  | [< 'Int i >] -> Some { eloc=(Lexing.dummy_pos, Lexing.dummy_pos);
			   econtent=Constant (IntCst (IInt, ICLiteral, Int64.of_int i)) }
  | [< _ >] -> None
  | [< >] -> None

let rec parse_simple_type = parser
  | [< s = ident >] -> Struct (s,TStandard)
  | [< 'Kwd "bool" >] -> Intermediate_format.Int IBool
  | [< 'Kwd "void" >] -> Void
  | [< 'Kwd "int"; s>] -> identify_int s
  | [< 'Kwd "char"; s >] -> identify_char s
  | [< 'Kwd "unsigned"; s >] -> identify_unsigned false s
  | [< 'Kwd "long"; s >] -> identify_long false s
  | [< 'Kwd "short"; s>] -> identify_short false s
  | [< 'Kwd "signed"; s >] -> identify_signed false s
  | [< 'Kwd "float"; s >] -> identify_float s
  | [< 'Kwd "double"; s >] -> identify_double s

(* Not really accurate (should take priorities and parentheses into account). *)
and modifier typ = parser
    [< 'Kwd "*"; s>] ->
      modifier (Pointer (PDataPointer { qualifier = []; plain_type = typ})) s
  | [< 'Kwd "&"; s>] ->
      modifier (LVReference (PDataPointer {qualifier=[];plain_type=typ})) s
  | [< 'Kwd "&&"; s>] ->
      modifier (RVReference (PDataPointer {qualifier=[];plain_type=typ})) s
  | [< 'Kwd "["; dim = opt_dimension; 'Kwd "]"; s >] ->
        modifier
          (Array { subtype = {qualifier=[];plain_type=typ}; 
                   dimension = dim }) s
  | [< >] -> typ

and template_args = parser
    [< t = template_arg; l = try_next_template_arg >] -> t :: l
  | [< >] -> []

and try_next_template_arg = parser
    [< 'Kwd ","; l = template_args >] -> l
  | [< >] -> []

(* Very incomplete *)
and template_arg = parser
  | [< 'Int i >] ->
      TPConstant (IntCst(IInt, ICLiteral, Int64.of_int i))
  | [< id = ident >] -> TPStructOrClass id
  | [< t = parse_type >] -> TPTypename t
      (* NB: we don't distinguish anymore between STA_TYPE and STA_ATOMIC*)
  | [< >] -> raise (Stream.Error "unsupported template instantiation")

and template t s =
  match s with parser
  | [< 'Kwd "<"; l = template_args; 'Kwd ">"; s >] ->
      maybe_qual (QTemplateInstance (t,l)) s
  | [< >] -> maybe_qual (QNamespace t) s

and ident = parser
    [< 'Ident s; st >] ->
      Frama_Clang_option.debug ~dkey "IDENT(%s)" s;
      template s st
  | [< 'Kwd "~"; 'Ident s >] -> 
      Frama_Clang_option.debug ~dkey "DESTR(%s)" s;
      { prequalification = []; decl_name = "~" ^ s }

and maybe_qual prefix = parser
    [< 'Kwd "::"; id >] ->
      Frama_Clang_option.debug ~dkey "QUAL";
      let s = ident id in
      { s with prequalification = prefix :: s.prequalification }
  | [< >] ->
      match prefix with
        | QNamespace s | QStructOrClass s ->
            { prequalification = []; decl_name = s }
        | QTemplateInstance (s,_) ->
            Frama_Clang_option.warning 
              "Dropping template arguments during parsing of identifier";
            { prequalification = []; decl_name = s } 

and parse_type_aux t = parser
    [< q = decl_flag; s>] -> parse_type_aux (add_qualifier q t) s
  | [< st = parse_simple_type; s >] ->
      let typ = modifier st s in { t with plain_type = typ }

and parse_type s =
  parse_type_aux { qualifier = []; plain_type = Void } s

let rec parse_args = parser
  | [< a = parse_type; l = maybe_args >] -> a :: l
and maybe_args = parser
  | [< 'Ident _; l = next_args >] -> l
  | [< l = next_args >] -> l
and next_args = parser
  | [< 'Kwd ","; l = parse_args >] -> l
  | [< 'Kwd ")" >] -> []

let parse_formals = parser
  | [< 'Kwd "("; args >] ->
    Frama_Clang_option.debug ~dkey "IS_SIG"; parse_args args

let plain_ident_or_func id = parser
  | [< f = ident; s >] ->
      let args = parse_formals s in
      f,
      Some 
        (FKFunction, 
         { result =
             { qualifier = [];
               plain_type = Struct (id, TStandard);
             };
           parameter = args;
           variadic = false;
         })
  | [< args = parse_formals >] ->
      let kind =
        if String.length id.decl_name > 0 && id.decl_name.[0] = '~' then
          FKDestructor true
        else begin
          match id.prequalification with
            | [] -> FKFunction
            | l ->
              match Extlib.last l with
                | QNamespace s | QStructOrClass s | QTemplateInstance (s,_)
                    when s = id.decl_name ->
                  FKConstructor true
                | QNamespace _ | QStructOrClass _ | QTemplateInstance _ ->
                  FKFunction
        end
      in
      (id,
       Some (kind, 
             { result = { qualifier = []; plain_type = Struct (id, TStandard)};
               parameter = args;
               variadic = false;
             }))
  | [<>] -> id, None

let parse_signature = parser
  | [< 'Kwd "::"; i = ident; s >] -> plain_ident_or_func i s
  | [< i = ident; s >] -> plain_ident_or_func i s
  | [< rt = parse_type; f = ident; s >] ->
      let args = parse_formals s in
      f, Some (FKFunction, { result = rt; parameter = args; variadic = false })

exception NoCxxName

let analyse_entrypoint s =
  try
    let input = Stream.of_string s in
    let input = lexer input in
    parse_signature input
  with
    | Stream.Error e -> 
      Frama_Clang_option.debug ~dkey
        "Unable to parse symbol %s as C++ function (%s). \
         Considering it as standard C identifier" s e;
      raise NoCxxName

let meth_name class_name tkind decl_name =
  let last_elt =
    match tkind with
      | TStandard -> QStructOrClass class_name.decl_name
      | TTemplateInstance i -> QTemplateInstance (class_name.decl_name, i)
  in
  let prequalification= List.append class_name.prequalification [last_elt] in
  { prequalification; decl_name }

let unqual_type t = { qualifier = []; plain_type = t }

let const_type t = { qualifier = [ Const ]; plain_type = t }

let obj_ptr t = { qualifier = []; plain_type = Pointer (PDataPointer t); }

let obj_lvref t = { qualifier = []; plain_type = LVReference (PDataPointer t); }

let obj_rvref t = { qualifier = []; plain_type = RVReference (PDataPointer t); }
